package org.jbake.render;

import org.apache.commons.configuration.CompositeConfiguration;
import org.jbake.app.ContentStore;
import org.jbake.app.Renderer;
import org.jbake.render.support.MockCompositeConfiguration;
import org.jbake.template.RenderingException;
import org.junit.Test;

import java.io.File;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.*;

public class TagsRendererTest {

    @Test
    public void returnsZeroWhenConfigDoesNotRenderTags() throws RenderingException {
        TagsRenderer renderer = new TagsRenderer();

        CompositeConfiguration compositeConfiguration = new MockCompositeConfiguration().withInnerBoolean(false);
        ContentStore contentStore = mock(ContentStore.class);

        Renderer mockRenderer = mock(Renderer.class);
        int renderResponse = renderer.render(mockRenderer, contentStore,
                new File("fake"), new File("fake"), compositeConfiguration);

        assertThat(renderResponse).isEqualTo(0);
    }

    @Test
    public void doesNotRenderWhenConfigDoesNotRenderTags() throws Exception {
        TagsRenderer renderer = new TagsRenderer();

        CompositeConfiguration compositeConfiguration = new MockCompositeConfiguration().withInnerBoolean(false);
        ContentStore contentStore = mock(ContentStore.class);
        Renderer mockRenderer = mock(Renderer.class);

        int renderResponse = renderer.render(mockRenderer, contentStore,
                new File("fake"), new File("fake"), compositeConfiguration);

        verify(mockRenderer, never()).renderTags(anyString());
    }

    @Test
    public void returnsOneWhenConfigRendersIndices() throws Exception {
        TagsRenderer renderer = new TagsRenderer();

        CompositeConfiguration compositeConfiguration = new MockCompositeConfiguration().withInnerBoolean(true);
        ContentStore contentStore = mock(ContentStore.class);
        Renderer mockRenderer = mock(Renderer.class);

        Set<String> tags = new HashSet(Arrays.asList("tag1", "tags2"));
        when(contentStore.getTags()).thenReturn(tags);

        when(mockRenderer.renderTags("random string")).thenReturn(1);

        int renderResponse = renderer.render(mockRenderer, contentStore,
                new File("fake"), new File("fake"), compositeConfiguration);

        assertThat(renderResponse).isEqualTo(1);
    }

    @Test
    public void doesRenderWhenConfigDoesNotRenderIndices() throws Exception {
        TagsRenderer renderer = new TagsRenderer();

        CompositeConfiguration compositeConfiguration = new MockCompositeConfiguration().withInnerBoolean(true);
        ContentStore contentStore = mock(ContentStore.class);
        Renderer mockRenderer = mock(Renderer.class);

        Set<String> tags = new HashSet(Arrays.asList("tag1", "tags2"));
        when(contentStore.getTags()).thenReturn(tags);

        int renderResponse = renderer.render(mockRenderer, contentStore,
                new File("fake"), new File("fake"), compositeConfiguration);

        verify(mockRenderer, times(1)).renderTags("random string");
    }

    @Test(expected = RenderingException.class)
    public void propogatesRenderingException() throws Exception {
        TagsRenderer renderer = new TagsRenderer();

        CompositeConfiguration compositeConfiguration = new MockCompositeConfiguration().withInnerBoolean(true);
        ContentStore contentStore = mock(ContentStore.class);
        Renderer mockRenderer = mock(Renderer.class);

        doThrow(new Exception()).when(mockRenderer).renderTags(anyString());

        int renderResponse = renderer.render(mockRenderer, contentStore,
                new File("fake"), new File("fake"), compositeConfiguration);

        verify(mockRenderer, never()).renderTags(anyString());
    }

}


